#include <kos.h>
#include "gfx/gfx.h"
#include "dc/filereq.h"

#include "gfx/bgpicraw.h"
#include "gfx/bgpicpal.h"

#include "gfx/fontpal.h"

#include "gfx/sintab.h"

#include <stdexcept>

#include "player.h"
#include "fformat.h"
#include "myendian.h"

emuEngine * myEmuEngine = 0;
sidTune * myTune = 0;
struct sidTuneInfo mySidInfo;
struct emuConfig myEmuConfig;
u16 sndbuffer[65536];
int errno;
u8 * screenbuf;
u32 palette[256];
BOOL paused;
int curSong;

#define COPYSCREENBUF \
	{ \
		int i; \
		for (i=0;i<320*240;i++) \
     	vram_l[i] = palette[screenbuf[i]]; \
  }

#define CLEARSCREENBUF \
	{ \
		int i; \
		for (i=0;i<320*240;i++) \
			screenbuf[i] = 0; \
	}


using namespace std;

void* operator new( size_t size ) throw(std::bad_alloc) {
	return ::malloc(size);
}

void* operator new[]( size_t size ) throw(std::bad_alloc) {
	return ::malloc(size);
}

void operator delete( void* todel ) throw() {
	::free(todel);
}

void operator delete[]( void* todel ) throw() {
	::free(todel);
}


void * snd_callback(int req)
	{
		if (paused)
			{
		  	int i;
		  	for (i=0;i<req>>1;++i)
		  		sndbuffer[i] = 0x8000;
			} else
			{
    		sidEmuFillBuffer ( *myEmuEngine,*myTune,sndbuffer, req );
    	}

		return sndbuffer;
	}

void SetPaletteEntry ( u8 i, u8 r, u8 g, u8 b )
	{
		palette[i] = (r<<16) | (g<<8) | b;
	}

static int fsize ( char * fname )
	{
		file_t f;
		int s;
		
		f = fs_open (fname, O_RDONLY);
		if (f==0)
			return -1;
		
		fs_seek (f, 0, SEEK_END);
		s=fs_tell(f);
		fs_close(f);
		
		return s;
	}

void PrintSongInfo(void)
	{
		char bla[256];
		int yp;

   	PrintStringCenter ( screenbuf, "SIDPLAY V0.01B", 0, 16 );
   	PrintStringCenter ( screenbuf, "2002,2003 BY CHN", 16, 16 );
   	
   	yp = 80;
   	yp+=PrintString ( screenbuf, mySidInfo.nameString, 0, yp, 16 );
   	yp+=PrintString ( screenbuf, mySidInfo.authorString, 0, yp, 16 );
   	yp+=PrintString ( screenbuf, mySidInfo.copyrightString, 0, yp, 16 );
   	yp+=16;
   	sprintf ( bla, "Subsong: %02d of %02d", curSong, mySidInfo.songs );
   	yp+=PrintString ( screenbuf, bla, 0, yp, 16 );
/*   	sprintf ( bla, "Free mem: %d", gm_availablesize() );
   	PrintStringCenter ( screenbuf, bla, 240-16, 16 );*/
	}

int main(int argc, char **argv)
	{
		int angle = 0;
    char bla[256];
    int i;
    BOOL sid_loaded = FALSE;
    u8 * sidmem = NULL;
    int sidsize = -1;
   	BOOL stereo = TRUE;

   	paused = FALSE;

    screenbuf = (u8*)malloc(320*240);
    memset((void*)screenbuf, 0, 320*240);

	  vid_set_mode(DM_320x240, PM_RGB888);

    SetPaletteEntry ( 128, 0, 0, 0 );
    SetPaletteEntry ( 129, 255, 255, 255 );
    SetPaletteEntry ( 130, 255, 0, 0 );
    SetPaletteEntry ( 131, 0, 255, 0 );
    SetPaletteEntry ( 132, 0, 0, 255 );
    SetPaletteEntry ( 133, 255, 255, 0 );
    SetPaletteEntry ( 134, 255, 96, 96 );

	  /* Picture */
	  for (i=0;i<16;i++)
	  	{
	    	u8 r = ((u16)bgpicpal[(i*3)+0]*24)/8;
	    	u8 g = ((u16)bgpicpal[(i*3)+1]*24)/8;
	    	u8 b = ((u16)bgpicpal[(i*3)+2]*24)/8;
	    	SetPaletteEntry ( i, r, g, b );
	  	}
	  /* Font */
	  for (i=0;i<8;i++)
	  	{
	    	u8 r = fontpal[(i*3)+0]<<2;
	    	u8 g = fontpal[(i*3)+1]<<2;
	    	u8 b = fontpal[(i*3)+2]<<2;
	    	SetPaletteEntry ( i+16, r, g, b );
	  	}

    myEmuEngine = new emuEngine();
    myEmuEngine->getConfig ( myEmuConfig );
    myEmuConfig.frequency = 44100;
    myEmuConfig.channels = SIDEMU_STEREO;
    myEmuConfig.bitsPerSample = SIDEMU_16BIT;
    myEmuConfig.emulateFilter = true;
 		myEmuConfig.sampleFormat = SIDEMU_SIGNED_PCM;

    snd_stream_init (snd_callback);

 		while (1)
 			{
	 			BOOL keypressed;
	    	u16 zoomsp=16,scale=0x0100;
				char fnamecopy[256];
				BOOL loaded = FALSE;

				while (!loaded)
					{
   					FSEL fsel;
   					file_t fh;
   					u8 * tempmem;
   					int fs;
						char fname[256];
						int retval;

   					fsel.fname = fname;
   					fsel.palette = palette;
	  				fsel.bgcolor = 128;
	  				fsel.filecolor = 129;
	  				fsel.dircolor = 133;
	  				fsel.curdircolor = 134;

	  				if (sid_loaded)
	  					{
		  					fsel.fnamepreselect = fnamecopy;
	  					} else
	  					{
		  					fsel.fnamepreselect = NULL;
	  					}

	  				vid_set_mode(DM_640x480, PM_RGB888);
	  				retval = FileRequest ( &fsel, "/", sid_loaded );
	  				vid_set_mode(DM_320x240, PM_RGB888);

	  				if (retval==0)
	  					{
	  						fs = fsize ( fname );
	  						tempmem = (u8*)malloc(fs);
	  						fh = fs_open ( fname, O_RDONLY );
	  						fs_read ( fh, tempmem, fs );
	  						fs_close ( fh );

	  						if (sid_loaded)
	  							{
		  							myTune->load ( tempmem, fs );
	  							} else
	  							{
		  							myTune = new sidTune ( tempmem, fs );
	  							}
	  				
    						if (!*myTune)
    							{	/* Failed loading the SID */
    								int k;
	    							free ( tempmem );
	    							if (sid_loaded)
	    								{
	    									myTune->load ( sidmem, sidsize );
	    									loaded = TRUE;
    									} else
	    								{
    										loaded = FALSE;
    									}
    								/* Display "Not a SID!" */
    								CLEARSCREENBUF;
   									PrintStringCenter ( screenbuf, "File is not a SID!", 120-8, 16 );
    								COPYSCREENBUF;
    								for (k=0;k<180;k++)
    									vid_waitvbl();
    							} else
    							{	/* Loaded the SID sucessfully */
										if ( sidmem!=NULL )
											free ( sidmem );

										sidmem = tempmem;
										sidsize = fs;
										strcpy ( fnamecopy, fname );
    								sid_loaded = loaded = TRUE;
    							}

    						myTune->getInfo ( mySidInfo );

	    					myEmuConfig.channels = stereo ? SIDEMU_STEREO : SIDEMU_MONO;
    						myEmuEngine->setConfig ( myEmuConfig );

    						sidEmuInitializeSong(*myEmuEngine,*myTune,curSong=1);
    						myEmuEngine->resetSecondsThisSong();
    					} else loaded = TRUE;
					}

				snd_stream_start(44100, stereo ? 1 : 0);

				keypressed = FALSE;
				while (1)
					{
						int i;
						ZoomStr zm;
						u8 cont;

						snd_stream_poll();

	    			/* Zoom Background */
    				zm.src = bgpicraw;
    				zm.dest = screenbuf;
    				zm.xpos = 0;
    				zm.ypos = 0;
    				zm.xscale = zm.yscale = 0x0008;
	    			if (scale<=0x0010 || scale>=0x0400)
	    				zoomsp = -zoomsp;
    				zm.xscale=zm.yscale=scale;
    				zm.xpos = -(160*zm.xscale) + (sinmul(angle, 0x8000)<<2);
    				zm.ypos = -(120*zm.yscale) + (cosmul(angle, 0x8000)<<2);
    				Zoom ( &zm );
   					angle+=16;
   					scale+=zoomsp;

						PrintSongInfo();

						vid_waitvbl();
						COPYSCREENBUF;
	           	
						/* Check Joypad */
						cont = maple_first_controller();
						if (cont)
							{
								cont_cond_t ctrl;

								while (cont_get_cond ( cont, &ctrl ) < 0);
								if (!(ctrl.buttons&CONT_START))
									break;
								else
								if ( (!(ctrl.buttons&CONT_DPAD_LEFT)) && !keypressed )
									{
										keypressed = TRUE;

		    						if (curSong>1)
		    							{
		    								curSong--;
		    								snd_stream_stop();
				    						sidEmuInitializeSong(*myEmuEngine,*myTune,curSong);
				    						myEmuEngine->resetSecondsThisSong();
				    						snd_stream_start(44100, stereo ? 1 : 0);
			    						}
									} else
								if ( (!(ctrl.buttons&CONT_DPAD_RIGHT)) && !keypressed )
									{
										keypressed = TRUE;
										
		    						if (curSong<mySidInfo.songs)
		    							{
		    								curSong++;
		    								snd_stream_stop();
				    						sidEmuInitializeSong(*myEmuEngine,*myTune,curSong);
				    						myEmuEngine->resetSecondsThisSong();
				    						snd_stream_start(44100, stereo ? 1 : 0);
			    						}
									} else
								if ( (!(ctrl.buttons&CONT_DPAD_UP)) && !keypressed )
									{
										keypressed = TRUE;
		    						snd_stream_stop();
		    						sidEmuInitializeSong(*myEmuEngine,*myTune,curSong);
		    						myEmuEngine->resetSecondsThisSong();
		    						snd_stream_start(44100, stereo ? 1 : 0);
									} else
								if ( (!(ctrl.buttons&CONT_X)) && !keypressed )
									{
										keypressed = TRUE;
										
										if ( stereo )
											{
												snd_stream_stop();
												stereo = FALSE;
					    					myEmuConfig.channels = stereo ? SIDEMU_STEREO : SIDEMU_MONO;
				    						myEmuEngine->setConfig ( myEmuConfig );
												snd_stream_start(44100, stereo ? 1 : 0);
											}
									} else
								if ( (!(ctrl.buttons&CONT_B)) && !keypressed )
									{
										keypressed = TRUE;
										
										if ( !stereo )
											{
												snd_stream_stop();
												stereo = TRUE;
					    					myEmuConfig.channels = stereo ? SIDEMU_STEREO : SIDEMU_MONO;
				    						myEmuEngine->setConfig ( myEmuConfig );
												snd_stream_start(44100, stereo ? 1 : 0);
											}
									} else
								/*if ( (!(ctrl.buttons&CONT_START)) && !keypressed )
									{
										keypressed = TRUE;
										paused = !paused;
									} else*/
									{
										keypressed = FALSE;
									}
							}
					}
				snd_stream_stop();
 			}

		return 0;
	}
